@@ -20,11 +20,14 @@ rjoin = lambda *paths: os.path.join(root_dir, *paths)
 data_dir = rjoin("data")
 licenses_json_path = rjoin(data_dir, "github_licenses.json")
 setup_tpl_path = rjoin(data_dir, "setup_tpl")
+pyproject_toml_tpl_path = rjoin(data_dir, "pyproject_toml_tpl.toml")
 gitignore_tpl_path = rjoin(data_dir, ".gitignore_tpl")
+gitattributes_tpl_path = rjoin(data_dir, ".gitattributes_tpl")
 github_ci_tpl_path = rjoin(data_dir, "github_ci_tpl.yml")
 gitlab_ci_tpl_path = rjoin(data_dir, "gitlab_ci_tpl.yml")
 github_ci_tpl_deploy_path = rjoin(data_dir, "github_ci_tpl_deploy.yml")
 github_ci_tpl_publish_path = rjoin(data_dir, "github_ci_tpl_publish.yml")
+github_ci_publish_2025_path = rjoin(data_dir, "github_ci_publish_2025.yml")
 
 pkg_dir = os.path.dirname(root_dir)
 pkg_join = lambda *paths: os.path.join(pkg_dir, *paths)
