@@ -38,23 +38,28 @@ from pprint import pprint
 from warnings import warn
 from typing import (
     Union,
-    Mapping,
-    Iterable,
-    Generator,
-    Sequence,
     Optional,
     Tuple,
     List,
     Literal,
 )
+from collections.abc import Mapping, Iterable, Generator, Sequence
 from configparser import ConfigParser
 from functools import partial
 
 import argh
 
 from wads.util import git
+from wads.toml_util import (
+    read_pyproject_toml,
+    write_pyproject_toml,
+    get_project_version,
+    set_project_version,
+    get_project_name,
+)
 
 CONFIG_FILE_NAME = "setup.cfg"
+PYPROJECT_FILE_NAME = "pyproject.toml"
 METADATA_SECTION = "metadata"
 OPTIONS_SECTION = "options"
 DFLT_OPTIONS = {
@@ -69,10 +74,35 @@ DOCSRC = "docsrc"
 DFLT_PUBLISH_DOCS_TO = None  # 'github'
 
 
+def _has_pyproject_toml(pkg_dir: str) -> bool:
+    """Check if package uses pyproject.toml"""
+    pkg_dir = os.path.realpath(pkg_dir)
+    if pkg_dir.endswith(os.sep):
+        pkg_dir = pkg_dir[:-1]
+    pyproject_path = os.path.join(pkg_dir, PYPROJECT_FILE_NAME)
+    return os.path.isfile(pyproject_path)
+
+
+def _has_setup_cfg(pkg_dir: str) -> bool:
+    """Check if package has setup.cfg"""
+    pkg_dir = os.path.realpath(pkg_dir)
+    if pkg_dir.endswith(os.sep):
+        pkg_dir = pkg_dir[:-1]
+    setup_cfg_path = os.path.join(pkg_dir, CONFIG_FILE_NAME)
+    return os.path.isfile(setup_cfg_path)
+
+
 def get_name_from_configs(pkg_dir, *, assert_exists=True):
-    """Get name from local setup.cfg (metadata section)"""
-    configs = read_configs(pkg_dir=pkg_dir)
-    name = configs.get("name", None)
+    """Get name from local config file (pyproject.toml or setup.cfg)"""
+    pkg_dir = _get_pkg_dir(pkg_dir)
+
+    # Prefer pyproject.toml if it exists
+    if _has_pyproject_toml(pkg_dir):
+        name = get_project_name(pkg_dir)
+    else:
+        configs = read_configs(pkg_dir=pkg_dir)
+        name = configs.get("name", None)
+
     if assert_exists:
         assert name is not None, "No name was found in configs"
     return name
@@ -87,7 +117,7 @@ def check_in(
     commit_message: str,
     *,
     work_tree: str = ".",
-    git_dir: Optional[str] = None,
+    git_dir: str | None = None,
     auto_choose_default_action: bool = False,
     bypass_docstring_validation: bool = False,
     bypass_tests: bool = False,
@@ -446,7 +476,7 @@ def _get_pkg_dir_and_name(pkg_dir, validate=True):
 
 def extract_pkg_dir_and_name(
     pkg_spec: PkgSpec, *, validate: bool = True
-) -> Tuple[str, str]:
+) -> tuple[str, str]:
     """
     Extracts the pkg_dir and pkg_dirname from the input `pkg_spec`.
     Optionally validates the pkg_dir is actually one (has a pkg_name/__init__.py file)
@@ -500,7 +530,7 @@ def extract_pkg_dir_and_name(
     return pkg_dir, pkg_name
 
 
-def folders_that_have_init_py_files(pkg_dir: PathStr) -> List[str]:
+def folders_that_have_init_py_files(pkg_dir: PathStr) -> list[str]:
     """
     Get a list of folders in the package directory that have an __init__.py file.
 
@@ -545,12 +575,17 @@ def current_configs(pkg_dir):
 
 def current_configs_version(pkg_dir):
     pkg_dir = _get_pkg_dir(pkg_dir)
-    return read_configs(pkg_dir=pkg_dir).get("version")
+
+    # Prefer pyproject.toml if it exists
+    if _has_pyproject_toml(pkg_dir):
+        return get_project_version(pkg_dir)
+    else:
+        return read_configs(pkg_dir=pkg_dir).get("version")
 
 
 # TODO: Both setup and twine are python. Change to use python objects directly.
 def update_setup_cfg(pkg_dir, *, new_deploy=False, version=None, verbose=True):
-    """Update setup.cfg.
+    """Update setup.cfg or pyproject.toml.
     If version is not given, will ask pypi (via http request) what the current version
     is, and increment that.
     """
@@ -562,16 +597,49 @@ def update_setup_cfg(pkg_dir, *, new_deploy=False, version=None, verbose=True):
     )
     pprint("\n{configs}\n")
     clog(verbose, pprint(configs))
-    write_configs(pkg_dir=pkg_dir, configs=configs)
+
+    # Write to pyproject.toml if it exists, otherwise setup.cfg
+    if _has_pyproject_toml(pkg_dir):
+        # Update pyproject.toml with the resolved configs
+        pyproject_data = read_pyproject_toml(pkg_dir)
+        if "project" not in pyproject_data:
+            pyproject_data["project"] = {}
+
+        # Map key fields back to pyproject.toml format
+        project = pyproject_data["project"]
+
+        if "version" in configs:
+            project["version"] = configs["version"]
+        if "description" in configs:
+            project["description"] = configs["description"]
+        if "name" in configs:
+            project["name"] = configs["name"]
+        if "url" in configs:
+            if "urls" not in project:
+                project["urls"] = {}
+            project["urls"]["Homepage"] = configs["url"]
+
+        write_pyproject_toml(pkg_dir, pyproject_data)
+    else:
+        write_configs(pkg_dir=pkg_dir, configs=configs)
 
 
 def set_version(pkg_dir, version):
-    """Update version setup.cfg"""
+    """Update version in config file (pyproject.toml and/or setup.cfg)"""
     pkg_dir = _get_pkg_dir(pkg_dir)
-    configs = read_configs(pkg_dir)
     assert isinstance(version, str), "version should be a string"
-    configs["version"] = version
-    write_configs(pkg_dir=pkg_dir, configs=configs)
+
+    # Update both files if both exist to keep them in sync
+    has_pyproject = _has_pyproject_toml(pkg_dir)
+    has_setup_cfg = _has_setup_cfg(pkg_dir)
+
+    if has_pyproject:
+        set_project_version(pkg_dir, version)
+
+    if has_setup_cfg:
+        configs = read_configs(pkg_dir)
+        configs["version"] = version
+        write_configs(pkg_dir=pkg_dir, configs=configs)
 
 
 def increment_configs_version(
@@ -579,27 +647,65 @@ def increment_configs_version(
     *,
     version=None,
 ):
-    """Increment version setup.cfg."""
+    """Increment version in config file (pyproject.toml and/or setup.cfg)."""
     pkg_dir = _get_pkg_dir(pkg_dir)
-    configs = read_configs(pkg_dir=pkg_dir)
-    version = _get_version(pkg_dir, version=version, configs=configs, new_deploy=False)
-    version = increment_version(version)
-    configs["version"] = version
-    write_configs(pkg_dir=pkg_dir, configs=configs)
+
+    # Check which config files exist
+    has_pyproject = _has_pyproject_toml(pkg_dir)
+    has_setup_cfg = _has_setup_cfg(pkg_dir)
+
+    # Determine the version to use
+    if version is None:
+        if has_pyproject:
+            # Get current version from pyproject.toml
+            current_version = get_project_version(pkg_dir)
+            version = increment_version(current_version) if current_version else "0.0.1"
+        else:
+            # Get version from setup.cfg
+            configs = read_configs(pkg_dir=pkg_dir)
+            version = _get_version(
+                pkg_dir, version=version, configs=configs, new_deploy=False
+            )
+            version = increment_version(version)
+
+    # Update both files if both exist to keep them in sync
+    if has_pyproject:
+        set_project_version(pkg_dir, version)
+
+    if has_setup_cfg:
+        configs = read_configs(pkg_dir)
+        configs["version"] = version
+        write_configs(pkg_dir=pkg_dir, configs=configs)
+
     return version
 
 
 def run_setup(pkg_dir):
-    """Run ``python setup.py sdist bdist_wheel``"""
-    print("--------------------------- setup_output ---------------------------")
+    """Run ``python -m build`` (modern PEP 517 compliant build)"""
+    print("--------------------------- build_output ---------------------------")
     pkg_dir = _get_pkg_dir(pkg_dir)
     original_dir = os.getcwd()
     os.chdir(pkg_dir)
-    setup_output = subprocess.run(
-        f"{sys.executable} setup.py sdist bdist_wheel".split(" ")
-    )
+
+    # Use python -m build (PEP 517 standard)
+    # Falls back to legacy setup.py if build module not available
+    try:
+        build_output = subprocess.run(
+            [sys.executable, "-m", "build"], capture_output=True, text=True
+        )
+        if build_output.returncode != 0:
+            print(f"Build stderr: {build_output.stderr}")
+            build_output.check_returncode()
+        print(build_output.stdout)
+    except (subprocess.CalledProcessError, FileNotFoundError) as e:
+        # Fallback to legacy method if build module not installed
+        print("Warning: 'build' module not found, falling back to setup.py")
+        print("Consider installing: pip install build")
+        build_output = subprocess.run(
+            f"{sys.executable} setup.py sdist bdist_wheel".split(" ")
+        )
+
     os.chdir(original_dir)
-    # print(f"{setup_output}\n")
 
 
 def twine_upload_dist(pkg_dir, *, options_str=None):
@@ -627,7 +733,7 @@ def twine_upload_dist(pkg_dir, *, options_str=None):
 
 # TODO: postprocess_ini_section_items and preprocess_ini_section_items: Add comma separated possibility?
 # TODO: Find out if configparse has an option to do this processing alreadys
-def postprocess_ini_section_items(items: Union[Mapping, Iterable]) -> Generator:
+def postprocess_ini_section_items(items: Mapping | Iterable) -> Generator:
     r"""Transform newline-separated string values into actual list of strings (assuming that intent)
 
     >>> section_from_ini = {
@@ -651,7 +757,7 @@ def postprocess_ini_section_items(items: Union[Mapping, Iterable]) -> Generator:
 
 
 # TODO: Find out if configparse has an option to do this processing already
-def preprocess_ini_section_items(items: Union[Mapping, Iterable]) -> Generator:
+def preprocess_ini_section_items(items: Mapping | Iterable) -> Generator:
     """Transform list values into newline-separated strings, in view of writing the value to a ini formatted section
 
     >>> section = {
@@ -688,10 +794,63 @@ def read_configs(
         pkg_dir, PathStr
     ), "It doesn't look like pkg_dir is a path. Did you perhaps invert pkg_dir and postproc order"
     pkg_dir = _get_pkg_dir(pkg_dir)
+
+    # Try pyproject.toml first (modern approach)
+    pyproject_filepath = pjoin(pkg_dir, PYPROJECT_FILE_NAME)
+    if os.path.isfile(pyproject_filepath):
+        pyproject_data = read_pyproject_toml(pkg_dir)
+        project_metadata = pyproject_data.get("project", {})
+
+        # Convert pyproject.toml format to setup.cfg format for compatibility
+        d = {}
+        if project_metadata:
+            # Map common fields
+            if "name" in project_metadata:
+                d["name"] = project_metadata["name"]
+            if "version" in project_metadata:
+                d["version"] = project_metadata["version"]
+            if "description" in project_metadata:
+                d["description"] = project_metadata["description"]
+            if "urls" in project_metadata:
+                urls = project_metadata["urls"]
+                if isinstance(urls, dict):
+                    # Try to get a root_url or homepage
+                    if "Homepage" in urls:
+                        d["url"] = urls["Homepage"]
+                    elif "repository" in urls:
+                        d["url"] = urls["repository"]
+                    elif "Repository" in urls:
+                        d["url"] = urls["Repository"]
+            # Handle authors
+            if "authors" in project_metadata:
+                authors = project_metadata["authors"]
+                if isinstance(authors, list) and authors:
+                    # Join author names
+                    d["author"] = ", ".join(
+                        a.get("name", "") for a in authors if a.get("name")
+                    )
+            if "license" in project_metadata:
+                license_info = project_metadata["license"]
+                if isinstance(license_info, dict):
+                    d["license"] = license_info.get("text", "")
+                else:
+                    d["license"] = str(license_info)
+            if "keywords" in project_metadata:
+                keywords = project_metadata["keywords"]
+                if isinstance(keywords, list):
+                    d["keywords"] = keywords
+            if "dependencies" in project_metadata:
+                d["install_requires"] = project_metadata["dependencies"]
+
+        if verbose:
+            print(f"Read configs from pyproject.toml: {d}")
+        return d
+
+    # Fall back to setup.cfg (legacy approach)
     config_filepath = pjoin(pkg_dir, CONFIG_FILE_NAME)
     c = ConfigParser()
     if os.path.isfile(config_filepath):
-        c.read_file(open(config_filepath, "r"))
+        c.read_file(open(config_filepath))
         if verbose:
             print(f"{section=}, {type(section)=}")
         try:
@@ -718,7 +877,7 @@ def write_configs(
     config_filepath = pjoin(pkg_dir, CONFIG_FILE_NAME)
     c = ConfigParser()
     if os.path.isfile(config_filepath):
-        c.read_file(open(config_filepath, "r"))
+        c.read_file(open(config_filepath))
 
     metadata_dict = dict(preproc(configs))
 
@@ -819,7 +978,7 @@ except ModuleNotFoundError:
     requests_is_installed = False
 
 
-def http_get_json(url, use_requests=requests_is_installed) -> Union[dict, None]:
+def http_get_json(url, use_requests=requests_is_installed) -> dict | None:
     """Make ah http request to url and get json, and return as python dict"""
 
     if use_requests:
@@ -854,9 +1013,9 @@ PYPI_PACKAGE_JSON_URL = "https://pypi.python.org/pypi/{package}/json"
 def versions_from_pypi(
     pkg_dir: PathStr,
     *,
-    name: Union[None, str] = None,
+    name: None | str = None,
     use_requests=requests_is_installed,
-) -> Union[str, None]:
+) -> str | None:
     """
     Return version of package on pypi.python.org using json.
 
@@ -866,8 +1025,12 @@ def versions_from_pypi(
     name = get_pkg_name(pkg_dir)
     url = PYPI_PACKAGE_JSON_URL.format(package=name)
 
-    pkg_info = http_get_json(url, use_requests=use_requests)
-    releases = pkg_info.get("releases", [])
+    try:
+        pkg_info = http_get_json(url, use_requests=use_requests)
+        releases = pkg_info.get("releases", [])
+    except Exception as e:
+        warn(f"Got an exception trying to get the versions from pypi: {e}")
+        return []
 
     # keep only the versions that don't have yanked=True
     def yanked_release(release_info_list):
@@ -880,9 +1043,9 @@ def versions_from_pypi(
 def highest_pypi_version(
     pkg_dir: PathStr,
     *,
-    name: Union[None, str] = None,
+    name: None | str = None,
     use_requests=requests_is_installed,
-) -> List[str]:
+) -> list[str]:
     """
     Return version of package on pypi.python.org using json.
 
@@ -900,7 +1063,7 @@ def highest_pypi_version(
 
 def current_pypi_version(
     pkg_dir: PathStr,
-) -> Union[str, None]:
+) -> str | None:
     """
     Return version of package on pypi.python.org using json.
 
@@ -921,7 +1084,7 @@ def current_pypi_version(
 
 def next_version_for_package(
     pkg_dir: PathStr,
-    name: Union[None, str] = None,
+    name: None | str = None,
     version_if_current_version_none="0.0.1",
     use_requests=requests_is_installed,
 ) -> str:
@@ -937,7 +1100,7 @@ def _get_version(
     pkg_dir: PathStr,
     version,
     configs,
-    name: Union[None, str] = None,
+    name: None | str = None,
     new_deploy=False,
 ):
     version = version or configs.get("version", None)
@@ -997,17 +1160,24 @@ def setup_cfg_version(pkg_spec: PkgSpec):
     return configs.get("version", None)
 
 
+def pyproject_toml_version(pkg_spec: PkgSpec):
+    """Get version from pyproject.toml file."""
+    pkg_dir, _ = extract_pkg_dir_and_name(pkg_spec)
+    return get_project_version(pkg_dir)
+
+
 def versions_from_different_sources(pkg_spec: PkgSpec):
     return {
         "tag": highest_tag_version(pkg_spec),
         "current_pypi": current_pypi_version(pkg_spec),
         "highest_not_yanked_pypi": highest_pypi_version(pkg_spec),
         "setup_cfg": setup_cfg_version(pkg_spec),
+        "pyproject_toml": pyproject_toml_version(pkg_spec),
     }
 
 
 ValidVersionSources = Literal[
-    "tag", "current_pypi", "highest_not_yanked_pypi", "setup_cfg"
+    "tag", "current_pypi", "highest_not_yanked_pypi", "setup_cfg", "pyproject_toml"
 ]
 
 
@@ -1032,9 +1202,10 @@ def validate_versions(versions: dict, action_when_not_valid=raise_error) -> dict
     #   Tip: Write the instructions in a github wiki/discussion/issue and provide link
 
     error_msg = ""
-    if versions["tag"] != versions["setup_cfg"]:
+    tag_version = versions.get("tag", None)
+    if tag_version is not None and tag_version != versions["setup_cfg"]:
         error_msg += (
-            f"Tag version ({versions['tag']}) is different "
+            f"Tag version ({tag_version}) is different "
             f"from setup.cfg's version: {versions['setup_cfg']}\n"
         )
     if versions["current_pypi"] != versions["highest_not_yanked_pypi"]:
@@ -1081,11 +1252,25 @@ def read_and_resolve_setup_configs(
     configs = read_configs(pkg_dir)
 
     # parse out name and root_url
-    assert (
-        "root_url" in configs or "url" in configs
-    ), "configs didn't have a root_url or url"
+    # If URL is missing, try to infer it from git
+    if "root_url" not in configs and "url" not in configs:
+        try:
+            from wads.populate import _get_pkg_url_from_pkg_dir
 
-    name = configs["name"] or pkg_dirname
+            git_url = _get_pkg_url_from_pkg_dir(pkg_dir)
+            configs["url"] = git_url
+            warn(
+                f"No url or root_url found in configs. Inferred from git: {git_url}. "
+                f"Consider adding it to your pyproject.toml or setup.cfg."
+            )
+        except Exception as e:
+            raise ValueError(
+                f"configs didn't have a root_url or url, and couldn't infer from git. "
+                f"Please add a url field to your pyproject.toml [project.urls] section "
+                f"or setup.cfg [metadata] section. Git error: {e}"
+            )
+
+    name = configs.get("name") or pkg_dirname
     if assert_names:
         assert (
             name == pkg_dirname
