@@ -3,6 +3,7 @@ import shutil
 import json
 from collections import ChainMap
 from urllib.parse import urlparse
+from warnings import warn
 
 # from functools import partial
 from typing import List, Optional
@@ -10,15 +11,23 @@ from wads import (
     pkg_join as wads_join,
     gitlab_ci_tpl_path,
     setup_tpl_path,
+    pyproject_toml_tpl_path,
     gitignore_tpl_path,
+    gitattributes_tpl_path,
     github_ci_tpl_deploy_path,
     github_ci_tpl_publish_path,
+    github_ci_publish_2025_path,
     wads_configs,
     wads_configs_file,
 )
 from wads.util import mk_conditional_logger, git, ensure_no_slash_suffix
 from wads.pack import write_configs
-from wads.licensing import license_body
+from wads.toml_util import write_pyproject_toml, read_pyproject_toml
+from wads.licensing import (
+    license_body,
+    resolve_author,
+    substitute_license_placeholders,
+)
 
 # from wads.pack_util import write_configs
 
@@ -29,7 +38,7 @@ populate_dflts = wads_configs.get(
     {
         "description": "There is a bit of an air of mystery around this project...",
         "root_url": None,
-        "author": None,
+        "author": os.environ.get("WADS_DFLT_AUTHOR"),
         "license": "mit",
         "description_file": "README.md",
         "long_description": "file:README.md",
@@ -55,27 +64,114 @@ def gen_readme_text(
 """
 
 
+def write_pyproject_configs(pkg_dir: str, configs: dict):
+    """
+    Write pyproject.toml file from template and configs.
+
+    Args:
+        pkg_dir: Path to package directory
+        configs: Dictionary of configuration values
+    """
+    import sys
+
+    if sys.version_info >= (3, 11):
+        import tomllib
+    else:
+        try:
+            import tomli as tomllib
+        except ImportError:
+            raise ImportError("tomli package required for Python < 3.11")
+
+    try:
+        import tomli_w
+    except ImportError:
+        raise ImportError("tomli_w package required for writing TOML")
+
+    # Read and parse the template as TOML
+    with open(pyproject_toml_tpl_path, "rb") as f:
+        data = tomllib.load(f)
+
+    # Prepare the data for substitution
+    name = configs.get("name", "mypackage")
+    version = configs.get("version", "0.0.1")
+    description = configs.get("description", "Package description")
+    url = configs.get("url", "")
+    license_name = configs.get("license", "mit")
+
+    # Warn if URL is missing or empty
+    if not url:
+        warn(
+            f"No URL provided for package '{name}'. "
+            f"PyPI requires a URL for package uploads. "
+            f"Please add 'url' or 'root_url' to your configuration."
+        )
+
+    # Update the parsed data with actual values
+    data["project"]["name"] = name
+    data["project"]["version"] = version
+    data["project"]["description"] = description
+
+    # Handle URLs - update homepage and repository
+    if url:
+        if "urls" not in data["project"]:
+            data["project"]["urls"] = {}
+        data["project"]["urls"]["Homepage"] = url
+        data["project"]["urls"]["Repository"] = url
+
+    # Update license using inline table syntax
+    data["project"]["license"] = {"text": license_name}
+
+    # Add optional fields if present
+    if configs.get("keywords"):
+        keywords = configs["keywords"]
+        if isinstance(keywords, str):
+            keywords = [k.strip() for k in keywords.split(",")]
+        data["project"]["keywords"] = keywords
+
+    if configs.get("author"):
+        authors = configs["author"]
+        if isinstance(authors, str):
+            data["project"]["authors"] = [{"name": authors}]
+        elif isinstance(authors, list):
+            data["project"]["authors"] = [
+                {"name": a} if isinstance(a, str) else a for a in authors
+            ]
+
+    if configs.get("install_requires"):
+        deps = configs["install_requires"]
+        if isinstance(deps, str):
+            deps = [d.strip() for d in deps.split(",") if d.strip()]
+        data["project"]["dependencies"] = deps
+
+    # Write the pyproject.toml
+    pyproject_path = os.path.join(pkg_dir, "pyproject.toml")
+    with open(pyproject_path, "wb") as f:
+        tomli_w.dump(data, f)
+
+
 # TODO: Function way to long -- break it up
 # TODO: Add a `defaults_from` in **configs that allows one to have several named defaults in wads_configs_file
 def populate_pkg_dir(
     pkg_dir,
     version: str = populate_dflts["version"],
     description: str = populate_dflts["description"],
-    root_url: Optional[str] = populate_dflts["root_url"],
-    author: Optional[str] = populate_dflts["author"],
+    root_url: str | None = populate_dflts["root_url"],
+    author: str | None = populate_dflts["author"],
     license: str = populate_dflts["license"],
     description_file: str = populate_dflts["description_file"],
-    keywords: Optional[List] = populate_dflts["keywords"],
-    install_requires: Optional[List] = populate_dflts["install_requires"],
+    keywords: list | None = populate_dflts["keywords"],
+    install_requires: list | None = populate_dflts["install_requires"],
     long_description=populate_dflts["long_description"],
     long_description_content_type=populate_dflts["long_description_content_type"],
     include_pip_install_instruction_in_readme=True,
     verbose: bool = populate_dflts["verbose"],
-    overwrite: List = (),
-    defaults_from: Optional[str] = None,
+    overwrite: list = (),
+    defaults_from: str | None = None,
     create_docsrc: bool = populate_dflts.get("create_docsrc", False),
     skip_docsrc_gen=False,
     skip_ci_def_gen=False,
+    migrate: bool = False,
+    create_gitattributes: bool = True,
     version_control_system=None,
     ci_def_path=None,
     ci_tpl_path=None,
@@ -112,6 +208,9 @@ def populate_pkg_dir(
     :param skip_docsrc_gen: Skip the generation of documentation stuff
     :param create_docsrc: If True, create and populate a `docsrc/` directory (overrides skip_docsrc_gen).
     :param skip_ci_def_gen: Skip the generation of the CI stuff
+    :param migrate: If True, migrate existing setup.cfg to pyproject.toml and old CI to new CI format.
+        Will fail if old CI has unmappable content.
+    :param create_gitattributes: If True (default), create a .gitattributes file with '*.ipynb linguist-documentation'.
     :param version_control_system: 'github' or 'gitlab' (will TRY to be resolved from root url if not given)
     :param ci_def_path: Path of the CI definition
     :param ci_tpl_path: Pater of the template definition
@@ -168,9 +267,12 @@ def populate_pkg_dir(
             "You should have a {name}/{name}/__init__.py structure. You don't."
         )
 
-    if os.path.isfile(pjoin("setup.cfg")):
-        with open(pjoin("setup.cfg"), "r"):
-            pass
+    # Check for existing config files (prioritize pyproject.toml over setup.cfg)
+    if os.path.isfile(pjoin("pyproject.toml")):
+        pass  # Will be handled later
+    elif os.path.isfile(pjoin("setup.cfg")):
+        _clog("... found existing setup.cfg (consider migrating to pyproject.toml)")
+        pass
 
     kwargs = dict(
         version=version,
@@ -207,28 +309,94 @@ def populate_pkg_dir(
         target_path = pjoin(resource_name)
         assert not os.path.isfile(target_path), f"{target_path} exists already"
         _clog(f"... making a {resource_name}")
-        with open(pjoin(resource_name), "wt") as fp:
+        with open(pjoin(resource_name), "w") as fp:
             fp.write(content)
 
     if should_update(".gitignore"):
         shutil.copy(gitignore_tpl_path, pjoin(".gitignore"))
 
+    if create_gitattributes and should_update(".gitattributes"):
+        _clog("... making a .gitattributes")
+        shutil.copy(gitattributes_tpl_path, pjoin(".gitattributes"))
+
     if project_type == "app":
         if should_update("requirements.txt"):
             with open(pjoin("requirements.txt"), "w") as f:
                 pass
 
     else:  # project_type == 'lib' or None
+        if should_update("pyproject.toml"):
+            setup_cfg_path = pjoin("setup.cfg")
+
+            # Check if setup.cfg exists (with or without migrate flag)
+            if os.path.isfile(setup_cfg_path):
+                _clog("... found setup.cfg, migrating to pyproject.toml")
+                from wads.migration import migrate_setuptools_to_hatching
+
+                try:
+                    # Migrate setup.cfg to pyproject.toml
+                    pyproject_content = migrate_setuptools_to_hatching(
+                        setup_cfg_path, defaults=configs
+                    )
+
+                    # Write the migrated content
+                    with open(pjoin("pyproject.toml"), "w") as f:
+                        f.write(pyproject_content)
+                except Exception as e:
+                    _clog(
+                        f"... migration failed ({e}), falling back to template-based creation"
+                    )
+                    if "pkg-dir" in configs:
+                        del configs["pkg-dir"]
+                    write_pyproject_configs(pjoin(""), configs)
+            else:
+                _clog("... making a 'pyproject.toml'")
+                if "pkg-dir" in configs:
+                    del configs["pkg-dir"]
+                write_pyproject_configs(pjoin(""), configs)
+        else:
+            # If pyproject.toml exists but URL is missing/empty, update it
+            if configs.get("url"):
+                from wads.toml_util import update_project_url, read_pyproject_toml
+
+                existing_data = read_pyproject_toml(pjoin(""))
+                existing_url = (
+                    existing_data.get("project", {}).get("urls", {}).get("Homepage", "")
+                )
+                if not existing_url:
+                    _clog(
+                        f"... updating URL in existing pyproject.toml to {configs['url']}"
+                    )
+                    update_project_url(pjoin(""), configs["url"])
+        # Keep setup.py for backward compatibility, but minimal
         if should_update("setup.py"):
             shutil.copy(setup_tpl_path, pjoin("setup.py"))
-        if should_update("setup.cfg"):
-            _clog("... making a 'setup.cfg'")
-            if "pkg-dir" in configs:
-                del configs["pkg-dir"]
-            write_configs(pjoin(""), configs)
 
     if should_update("LICENSE"):
         _license_body = license_body(configs["license"])
+
+        # Resolve author for license placeholder substitution
+        # Try to get authors from existing pyproject.toml if it exists
+        pyproject_authors = None
+        if os.path.isfile(pjoin("pyproject.toml")):
+            try:
+                pyproject_data = read_pyproject_toml(pjoin(""))
+                pyproject_authors = pyproject_data.get("project", {}).get("authors")
+            except Exception:
+                pass  # If reading fails, continue without pyproject authors
+
+        resolved_author = resolve_author(
+            author=configs.get("author"),
+            pyproject_authors=pyproject_authors,
+            url=configs.get("url"),
+        )
+
+        # Substitute placeholders in license
+        _license_body = substitute_license_placeholders(
+            _license_body,
+            author=resolved_author,
+        )
+
         save_txt_to_pkg("LICENSE", _license_body)
 
     if should_update("README.md"):
@@ -264,9 +432,37 @@ def populate_pkg_dir(
             assert name in ci_def_path and name in _get_pkg_url_from_pkg_dir(
                 pkg_dir
             ), f"The name wasn't found in both the ci_def_path AND the git url, so I'm going to be safe and do nothing"
-            # TODO: Handle user_email better than this and make more visible
-            user_email = kwargs.get("user_email", "thorwhalen1@gmail.com")
-            _add_ci_def(ci_def_path, ci_tpl_path, root_url, name, _clog, user_email)
+
+            # Check if we should migrate old CI to new format
+            if migrate and version_control_system == "github":
+                old_ci_path = pjoin(".github/workflows/ci.yml")
+                if os.path.isfile(old_ci_path):
+                    _clog(f"... migrating old CI from {old_ci_path} to {ci_def_path}")
+                    from wads.migration import migrate_github_ci_old_to_new
+
+                    try:
+                        new_ci_content = migrate_github_ci_old_to_new(
+                            old_ci_path, defaults={"project_name": name}
+                        )
+                        os.makedirs(os.path.dirname(ci_def_path), exist_ok=True)
+                        with open(ci_def_path, "w") as f:
+                            f.write(new_ci_content)
+                        _clog(f"... successfully migrated CI to {ci_def_path}")
+                    except Exception as e:
+                        raise RuntimeError(
+                            f"Failed to migrate CI file {old_ci_path}: {e}\n"
+                            f"The old CI may contain configurations that cannot be automatically migrated."
+                        ) from e
+                else:
+                    # No old CI to migrate, create new one from template
+                    user_email = kwargs.get("user_email", "thorwhalen1@gmail.com")
+                    _add_ci_def(
+                        ci_def_path, ci_tpl_path, root_url, name, _clog, user_email
+                    )
+            else:
+                # Not migrating or not github, use template
+                user_email = kwargs.get("user_email", "thorwhalen1@gmail.com")
+                _add_ci_def(ci_def_path, ci_tpl_path, root_url, name, _clog, user_email)
 
     return name
 
@@ -275,10 +471,14 @@ _unknown_version_control_system = object()
 
 
 def _ensure_url_from_url_root_and_name(configs: dict):
-    if "url" not in configs and "url_root" in configs and "name" in configs:
-        if not configs["url_root"].endswith("/"):
-            configs["urls_root"] += "/"
-        configs["url"] = configs["urls_root"] + configs["name"]
+    """Ensure configs has a 'url' field, constructing it from root_url/url_root if needed."""
+    # Handle both 'root_url' and 'url_root' for backward compatibility
+    root_url = configs.get("root_url") or configs.get("url_root")
+
+    if "url" not in configs and root_url and "name" in configs:
+        if not root_url.endswith("/"):
+            root_url += "/"
+        configs["url"] = root_url + configs["name"]
     return configs
 
 
@@ -307,7 +507,7 @@ def _resolve_ci_def_and_tpl_path(
                 ci_tpl_path = github_ci_tpl_deploy_path
             else:  # project_type == 'lib', etc.
                 print(f"project_type is {project_type}")
-                ci_tpl_path = github_ci_tpl_publish_path
+                ci_tpl_path = github_ci_publish_2025_path
         elif version_control_system == "gitlab":
             ci_tpl_path = gitlab_ci_tpl_path
         else:
@@ -497,14 +697,22 @@ def populate_proj_from_url(
         proj_rootdir = os.path.join(proj_rootdir, proj_root_dir_for_name[url_name])
     _clog(f"proj_rootdir={proj_rootdir}")
 
+    proj_path = os.path.join(proj_rootdir, proj_name)
+
     with cd(proj_rootdir):
-        _clog(f"cloning {url}...")
-        subprocess.check_output(f"git clone {url}", shell=True).decode()
+        # Only clone if the directory doesn't exist
+        if not os.path.isdir(proj_path):
+            _clog(f"cloning {url}...")
+            subprocess.check_output(f"git clone {url}", shell=True).decode()
+        else:
+            _clog(f"Directory {proj_path} already exists, skipping clone...")
+
         _clog(f"populating package folder...")
         populate_pkg_dir(
-            os.path.join(proj_rootdir, proj_name),
+            proj_path,
             defaults_from=url_name,
             description=description,
+            root_url=root_url,  # Pass root_url explicitly
             license=license,
             **kwargs,
         )
